@@ -1,12 +1,14 @@
-import time
+import json
 import logging
-from cgi import escape
+import pickletools
+import time
+import traceback
 
 from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile
 from zope.app.publication.zopepublication import ZopePublication, Cleanup
 from zope.publisher.browser import BrowserView
 from zope.publisher.interfaces.browser import IBrowserRequest
-from zope.component import adapts, queryUtility
+from zope.component import adapter, queryUtility
 from zope.interface import Interface
 from zope.security.proxy import removeSecurityProxy
 from zope.cachedescriptors.property import Lazy
@@ -18,14 +20,14 @@ from ZODB.POSException import POSKeyError
 from persistent import Persistent
 from persistent.TimeStamp import TimeStamp
 import transaction
-import simplejson
 
 from zodbbrowser import __version__, __homepage__
-from zodbbrowser.history import ZodbObjectHistory
-from zodbbrowser.interfaces import IValueRenderer
-from zodbbrowser.interfaces import IDatabaseHistory
-from zodbbrowser.state import ZodbObjectState
+from zodbbrowser.compat import escape, StringIO, BytesIO
 from zodbbrowser.diff import compareDictsHTML
+from zodbbrowser.history import ZodbObjectHistory
+from zodbbrowser.interfaces import IDatabaseHistory
+from zodbbrowser.interfaces import IValueRenderer
+from zodbbrowser.state import ZodbObjectState
 from zodbbrowser.value import pruneTruncations, TRUNCATIONS
 
 
@@ -85,6 +87,15 @@ class VeryCarefulView(BrowserView):
                 raise Exception("ZODB connection not available for this request")
             return obj._p_jar
 
+    def findClosestPersistent(self):
+        obj = removeSecurityProxy(self.context)
+        while not isinstance(obj, Persistent):
+            try:
+                obj = obj.__parent__
+            except AttributeError:
+                return None
+        return obj
+
     @property
     def readonly(self):
         return self.jar.isReadOnly()
@@ -107,11 +118,10 @@ class VeryCarefulView(BrowserView):
                 transaction.abort()
 
 
+@adapter(Interface, IBrowserRequest)
 class ZodbInfoView(VeryCarefulView):
     """Zodb browser view"""
 
-    adapts(Interface, IBrowserRequest)
-
     template = ViewPageTemplateFile('templates/zodbinfo.pt')
     confirmation_template = ViewPageTemplateFile('templates/confirm_rollback.pt')
 
@@ -142,8 +152,8 @@ class ZodbInfoView(VeryCarefulView):
             self.requestedState = self._tidToTimestamp(rtid)
             if self.request.get('confirmed') == '1':
                 self.history.rollback(rtid)
-                transaction.get().note('Rollback to old state %s'
-                                        % self.requestedState)
+                transaction.get().note(u'Rollback to old state %s'
+                                       % self.requestedState)
                 self.made_changes = True
                 self._redirectToSelf()
                 return ''
@@ -182,15 +192,6 @@ class ZodbInfoView(VeryCarefulView):
                 raise UserError('There is no object with OID 0x%x' % oid)
         return obj
 
-    def findClosestPersistent(self):
-        obj = removeSecurityProxy(self.context)
-        while not isinstance(obj, Persistent):
-            try:
-                obj = obj.__parent__
-            except AttributeError:
-                return None
-        return obj
-
     def getRequestedTid(self):
         if 'tid' in self.request:
             return self.request['tid']
@@ -233,7 +234,7 @@ class ZodbInfoView(VeryCarefulView):
         return u64(root._p_oid)
 
     def locate_json(self, path): # AJAX view
-        return simplejson.dumps(self.locate(path))
+        return json.dumps(self.locate(path))
 
     def truncated_ajax(self, id): # AJAX view
         return TRUNCATIONS.get(id)
@@ -275,7 +276,7 @@ class ZodbInfoView(VeryCarefulView):
                 continue
             if not here.endswith('/'):
                 here += '/'
-            here += step.encode('utf-8')
+            here += step
             try:
                 child = obj[step]
             except Exception:
@@ -343,11 +344,28 @@ class ZodbInfoView(VeryCarefulView):
         for name, url in self.getBreadcrumbs():
             if url:
                 html.append('<a href="%s">%s</a>' % (escape(url, True),
-                                                     escape(name)))
+                                                     escape(name, False)))
             else:
-                html.append(escape(name))
+                html.append(escape(name, False))
         return ''.join(html)
 
+    def getDisassembledPickleData(self):
+        pickle = BytesIO(self.state.pickledState)
+        out = StringIO()
+        memo = {}
+        # 1st pickle: the class
+        try:
+            pickletools.dis(pickle, out, memo)
+        except Exception as e:
+            out.write(''.join(traceback.format_exception_only(type(e), e)))
+        # 2st pickle: actual instance data
+        out.write('\n')
+        try:
+            pickletools.dis(pickle, out, memo)
+        except Exception as e:
+            out.write(''.join(traceback.format_exception_only(type(e), e)))
+        return out.getvalue()
+
     def listAttributes(self):
         attrs = self.state.listAttributes()
         if attrs is None:
@@ -370,7 +388,7 @@ class ZodbInfoView(VeryCarefulView):
                                          _history=self.history)
                 state = interp.asDict()
                 error = interp.getError()
-            except Exception, e:
+            except Exception as e:
                 state = {}
                 error = '%s: %s' % (e.__class__.__name__, e)
             results.append(dict(state=state, error=error))
@@ -382,10 +400,15 @@ class ZodbInfoView(VeryCarefulView):
         state = self._loadHistoricalState()
         results = []
         for n, d in enumerate(self.history):
-            short = (str(time.strftime('%Y-%m-%d %H:%M:%S',
-                                       time.localtime(d['time']))) + " "
-                     + d['user_name'] + " "
-                     + d['description'])
+            utc_timestamp = str(time.strftime('%Y-%m-%d %H:%M:%S',
+                                              time.gmtime(d['time'])))
+            local_timestamp = str(time.strftime('%Y-%m-%d %H:%M:%S',
+                                                time.localtime(d['time'])))
+            try:
+                user_location, user_id = d['user_name'].split()
+            except ValueError:
+                user_location = None
+                user_id = d['user_name']
             url = self.getUrl(tid=u64(d['tid']))
             current = (d['tid'] == self.state.tid and
                        self.state.requestedTid is not None)
@@ -393,10 +416,13 @@ class ZodbInfoView(VeryCarefulView):
             oldState = state[n + 1]['state']
             diff = compareDictsHTML(curState, oldState, d['tid'])
 
-            results.append(dict(short=short, utid=u64(d['tid']),
+            results.append(dict(utid=u64(d['tid']),
                                 href=url, current=current,
                                 error=state[n]['error'],
-                                diff=diff, **d))
+                                diff=diff, user_id=user_id,
+                                user_location=user_location,
+                                utc_timestamp=utc_timestamp,
+                                local_timestamp=local_timestamp, **d))
 
         # number in reverse order
         for i in range(len(results)):
@@ -405,16 +431,15 @@ class ZodbInfoView(VeryCarefulView):
         return results
 
     def _tidToTimestamp(self, tid):
-        if isinstance(tid, str) and len(tid) == 8:
+        if isinstance(tid, bytes) and len(tid) == 8:
             return str(TimeStamp(tid))
         return tid_repr(tid)
 
 
+@adapter(Interface, IBrowserRequest)
 class ZodbHistoryView(VeryCarefulView):
     """Zodb history view"""
 
-    adapts(Interface, IBrowserRequest)
-
     template = ViewPageTemplateFile('templates/zodbhistory.pt')
 
     version = __version__
@@ -422,12 +447,13 @@ class ZodbHistoryView(VeryCarefulView):
 
     page_size = 5
 
-    def render(self):
+    def update(self):
         self._started = time.time()
         pruneTruncations()
         if 'page_size' in self.request:
             self.page_size = max(1, int(self.request['page_size']))
         self.history = IDatabaseHistory(self.jar)
+        self.request.hold(Cleanup(self.history.cleanup))
         if 'page' in self.request:
             self.page = int(self.request['page'])
         elif 'tid' in self.request:
@@ -440,6 +466,9 @@ class ZodbHistoryView(VeryCarefulView):
             self.page = self.last_page
         self.last_idx = max(0, len(self.history) - self.page * self.page_size)
         self.first_idx = max(0, self.last_idx - self.page_size)
+
+    def render(self):
+        self.update()
         return self.template()
 
     def renderingTime(self):
@@ -470,9 +499,26 @@ class ZodbHistoryView(VeryCarefulView):
         results = []
         for n, d in enumerate(self.history[self.first_idx:self.last_idx]):
             utid = u64(d.tid)
-            short = "%s %s %s" % (TimeStamp(d.tid),
-                                  d.user,
-                                  d.description)
+            ts = TimeStamp(d.tid).timeTime()
+            utc_timestamp = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(ts))
+            local_timestamp = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(ts))
+            try:
+                user_location, user_id = d.user.split()
+            except ValueError:
+                user_location = None
+                user_id = d.user
+            if isinstance(user_location, bytes):
+                user_location = user_location.decode('UTF-8', 'replace')
+            if isinstance(user_id, bytes):
+                user_id = user_id.decode('UTF-8', 'replace')
+            description = d.description
+            if isinstance(description, bytes):
+                description = description.decode('UTF-8', 'replace')
+            try:
+                size = d._tend - d._tpos
+            except AttributeError:
+                size = None
+            ext = d.extension if isinstance(d.extension, dict) else {}
             objects = []
             for record in d:
                 obj = self.jar.get(record.oid)
@@ -490,15 +536,23 @@ class ZodbHistoryView(VeryCarefulView):
                 summary = '1 object record'
             else:
                 summary = '%d object records' % len(objects)
+            if size is not None:
+                summary += ' (%d bytes)' % size
             results.append(dict(
                 index=(self.first_idx + n + 1),
-                short=short,
+                utc_timestamp=utc_timestamp,
+                local_timestamp=local_timestamp,
+                user_id=user_id,
+                user_location=user_location,
+                description=description,
                 utid=utid,
                 current=(d.tid == requested_tid),
                 href=self.getUrl(tid=utid),
+                size=size,
                 summary=summary,
                 hidden=(len(objects) > 5),
                 objects=objects,
+                **ext
             ))
         if results and not requested_tid and self.page == 0:
             results[-1]['current'] = True
