@@ -1,13 +1,15 @@
-import os
+import doctest
 import gc
+import glob
+import json
+import logging
+import os
 import re
+import shutil
 import sys
 import tempfile
-import shutil
-import doctest
-import unittest
 import threading
-from cgi import escape
+import unittest
 
 import transaction
 from lxml.html import fromstring, tostring
@@ -16,17 +18,59 @@ from ZODB.POSException import ReadOnlyError
 from ZODB.FileStorage.FileStorage import FileStorage
 from ZODB.DB import DB
 from zope.testing.renormalizing import RENormalizing
-from zope.testbrowser.browser import Browser
+from zope.testbrowser.browser import Browser as _Browser, HTTPError
 from zope.testbrowser.interfaces import IBrowser
 from zope.app.testing import setup
 from zope.app.publication.zopepublication import ZopePublication
 from zope.app.folder.folder import Folder
 from zope.app.appsetup.interfaces import DatabaseOpened
 from zope.app.appsetup.bootstrap import bootStrapSubscriber
-from zope.interface import Interface, implementsOnly
+from zope.interface import Interface, implementer_only
 
-from zodbbrowser.standalone import main, serve_forever, stop_serving
 from zodbbrowser import standalone
+from zodbbrowser.compat import basestring, StringIO, escape
+from zodbbrowser.standalone import main, serve_forever, stop_serving
+from zodbbrowser.value import resetTruncations
+
+
+class InternalServerError(Exception):
+
+    def __init__(self, url, log):
+        super(InternalServerError, self).__init__("%s\n%s" % (url, log))
+        self.url = url
+        self.log = log
+
+
+class Browser(_Browser):
+
+    capture_log = 'SiteError'  # keep '' to capture everything
+    log_format = '%(message)s' # '%(name)s %(levelname)s %(message)s' also useful
+    log_level = logging.DEBUG
+
+    @property
+    def json(self):
+        return json.loads(self.contents)
+
+    # XXX: we should also wrap Form.submit the same way
+
+    def open(self, url, *args, **kwargs):
+        buffer = StringIO()
+        logger = logging.getLogger(self.capture_log)
+        handler = logging.StreamHandler(buffer)
+        handler.setFormatter(logging.Formatter(self.log_format))
+        level = logger.level
+        logger.addHandler(handler)
+        try:
+            logger.setLevel(self.log_level)
+            return super(Browser, self).open(url, *args, **kwargs)
+        except HTTPError as e:
+            if e.code == 500:
+                raise InternalServerError(url, buffer.getvalue())
+            else:
+                raise
+        finally:
+            logger.handlers.remove(handler)
+            logger.setLevel(level)
 
 
 class ServerController(object):
@@ -64,7 +108,6 @@ class ServerController(object):
             self.server_thread = None
             self.url = None
             self.port_number = None
-            # XXX: leaves zserver threads behind
             # XXX: leaves global Zope state, you'll need to run
             #      setup.placelessTearDown()
 
@@ -80,8 +123,9 @@ class IMyOwnInterface(Interface):
     pass
 
 
+@implementer_only(IMyOwnInterface)
 class PersistentSubclassThatUsesImplementsOnly(Persistent):
-    implementsOnly(IMyOwnInterface)
+    """See https://bugs.launchpad.net/zodbbrowser/+bug/1185175."""
 
 
 class TestsWithServer(object):
@@ -111,12 +155,15 @@ class TestsWithServer(object):
         connection = db.open()
         root = connection.root()
         root_folder = root[ZopePublication.root_name]
+        transaction.get().note(u"setUp creating root folder")
+        transaction.commit()
         # This is not a great way to set up test fixtures, but it'll do
         # for now
         cls.createTestDataForBrowsing(root_folder)
         cls.createTestDataForRollbacking(root_folder)
         cls.createTestDataForRollbackCanBeCancelled(root_folder)
         cls.createTestDataForImplementsOnly(root_folder)
+        cls.createTestDataForTruncation(root_folder)
         connection.close()
         db.close()
 
@@ -124,22 +171,27 @@ class TestsWithServer(object):
     def createTestDataForBrowsing(cls, root_folder):
         # set up data that browsing.txt expects
         root_folder['browsing'] = Folder()
+        transaction.get().note(u"createTestDataForBrowsing")
         transaction.commit()
 
     @classmethod
     def createTestDataForRollbacking(cls, root_folder):
         # set up data that rollbacking.txt expects
         root_folder['rollbacking'] = Folder()
+        transaction.get().note(u"createTestDataForRollbacking (1)")
         transaction.commit()
         root_folder['rollbacking'].random_attribute = 'hey'
+        transaction.get().note(u"createTestDataForRollbacking (2)")
         transaction.commit()
 
     @classmethod
     def createTestDataForRollbackCanBeCancelled(cls, root_folder):
         # set up data that rollback-can-be-cancelled.txt expects
         root_folder['rbcbc'] = Folder()
+        transaction.get().note(u"createTestDataForRollbackCanBeCancelled (1)")
         transaction.commit()
         root_folder['rbcbc'].random_attribute = 'hey'
+        transaction.get().note(u"createTestDataForRollbackCanBeCancelled (2)")
         transaction.commit()
 
     @classmethod
@@ -147,6 +199,22 @@ class TestsWithServer(object):
         # set up data that implements-only.txt expects
         root_folder['io'] = Folder()
         root_folder['io'].crash = PersistentSubclassThatUsesImplementsOnly()
+        transaction.get().note(u"createTestDataForImplementsOnly")
+        transaction.commit()
+
+    @classmethod
+    def createTestDataForTruncation(cls, root_folder):
+        # set up data that truncation.txt expects
+        root_folder['longvalue'] = Folder()
+        root_folder['longvalue'].long_attribute = u'''
+            This is an attribute that has a rather long value.
+            Because these happen sometimes in real life, and
+            to display the entire very long string would be
+            rather painful.  So I'm adding some text here, not
+            too much, but enough for the truncation logic to kick
+            in.  Is this enough?  I hope so.
+        '''
+        transaction.get().note(u"createTestDataForTruncation")
         transaction.commit()
 
 
@@ -178,10 +246,14 @@ class TestCanCreateEmptyDataFs(unittest.TestCase):
         self.assertTrue('persistent.mapping.PersistentMapping' in browser.contents)
 
     def test_cannot_start_in_read_only_mode(self):
-        self.assertRaises(ReadOnlyError, self.server.run, self.empty_fs)
-        # Due to a bug in ZODB, the new database *is* created, it just
-        # has no objects in it
-        # self.assertTrue(not os.path.exists(self.empty_fs))
+        with self.assertRaises((IOError, ReadOnlyError)) as cm:
+            self.server.run(self.empty_fs)
+        if isinstance(cm.exception, ReadOnlyError):  # pragma: nocover
+            # Due to a bug in older versions of ZODB, the new database *was*
+            # created, it just had no objects in it.
+            pass
+        else:
+            self.assertTrue(not os.path.exists(self.empty_fs))
 
 
 def printXPath(html, xpath, pretty_print=True):
@@ -221,7 +293,9 @@ def printResults(html, method, arg, pretty_print=True):
         # it would be nice to extract the charset from the content-type
         # header, but let's assume UTF-8, which is the only charset that
         # we use in our system
-        html = html.contents.decode('UTF-8')
+        html = html.contents
+    if isinstance(html, bytes):
+        html = html.decode('UTF-8')
     # XXX: not the most appropriate place for this.  I cannot do it with
     # a renormalizer since at that point the server URL is not yet known
     html = html.replace(TestsWithServer.url, 'http://localhost/')
@@ -235,9 +309,11 @@ def printResults(html, method, arg, pretty_print=True):
                 fixupWhitespace(element)
             value = tostring(element, pretty_print=pretty_print).rstrip()
         if value:
-            print value
+            if not isinstance(value, str):
+                value = value.decode('UTF-8')
+            print(value)
     if not results:
-        print "Not found: %s" % arg
+        print("Not found: %s" % arg)
 
 
 def stripify(s):
@@ -272,7 +348,7 @@ def fixupWhitespace(element, indent=0, step=2, split_if_longer=38):
     element.text = stripify(element.text)
     # heuristic for splitting long elements
     should_split = (len(str(element.attrib)) > split_if_longer or
-                    len(escape(element.text)) > split_if_longer)
+                    len(escape(element.text, False)) > split_if_longer)
     if should_split and element.text:
         element.text = ('\n' + ' ' * (indent + step) + element.text.lstrip())
     if children:
@@ -296,6 +372,7 @@ def fixupWhitespace(element, indent=0, step=2, split_if_longer=38):
 
 
 def setUp(test):
+    resetTruncations()
     test.globs['Browser'] = Browser
     test.globs['printXPath'] = printXPath
     test.globs['printCSSPath'] = printCSSPath
@@ -322,14 +399,15 @@ def test_suite():
         # zope.container 4.0.0 made Folder objects inherit from BTreeContainer
         # this adds one new attribute that our tests don't expect to see
         (re.compile(r'<strong>_BTreeContainer__len</strong>'), ''),
+        # Python 3 has unicode strings without u prefixes
+        (re.compile(r"u('.*')"), r'\1'),
+        (re.compile(r": u'$", re.MULTILINE), ": '"),
     ])
     optionflags = (doctest.REPORT_ONLY_FIRST_FAILURE |
                    doctest.REPORT_NDIFF | doctest.NORMALIZE_WHITESPACE)
     here = os.path.dirname(__file__)
-    for filename in sorted(os.listdir(here)):
-        if not filename.endswith('.txt') or filename.startswith('.'):
-            continue
-        test = doctest.DocFileSuite(filename,
+    for filename in sorted(glob.glob(os.path.join(here, '*.txt'))):
+        test = doctest.DocFileSuite(os.path.basename(filename),
                                     setUp=setUp,
                                     checker=checker,
                                     optionflags=optionflags)
