@@ -1,21 +1,24 @@
-import inspect
-
 from ZODB.utils import tid_repr
-from ZODB.interfaces import IConnection
+from ZODB.interfaces import IConnection, IStorageIteration
 from persistent import Persistent
 from zope.proxy import removeAllProxies
-from zope.interface import implements
-from zope.component import adapts
+from zope.interface import implementer
+from zope.component import adapter
 
 from zodbbrowser.interfaces import IObjectHistory, IDatabaseHistory
 from zodbbrowser import cache
 
+try:
+    from ZODB.mvccadapter import MVCCAdapterInstance
+except ImportError:  # pragma: no-cover
+    class MVCCAdapterInstance(object):
+        """Placeholder so we can register an adapter that will not be used."""
 
+
+@adapter(Persistent)
+@implementer(IObjectHistory)
 class ZodbObjectHistory(object):
 
-    adapts(Persistent)
-    implements(IObjectHistory)
-
     def __init__(self, obj):
         self._obj = removeAllProxies(obj)
         self._connection = self._obj._p_jar
@@ -43,18 +46,7 @@ class ZodbObjectHistory(object):
         See the 'history' method of ZODB.interfaces.IStorage.
         """
         size = 999999999999 # "all of it"; ought to be sufficient
-        # NB: ClientStorage violates the interface by calling the last
-        # argument 'length' instead of 'size'.  To avoid problems we must
-        # use positional argument syntax here.
-        # NB: FileStorage in ZODB 3.8 has a mandatory second argument 'version'
-        # FileStorage in ZODB 3.9 doesn't accept a 'version' argument at all.
-        # This check is ugly, but I see no other options if I want to support
-        # both ZODB versions :(
-        if 'version' in inspect.getargspec(self._storage.history)[0]:
-            version = None
-            self._history = self._storage.history(self._oid, version, size)
-        else:
-            self._history = self._storage.history(self._oid, size=size)
+        self._history = self._storage.history(self._oid, size=size)
         self._index_by_tid()
 
     def _index_by_tid(self):
@@ -64,7 +56,12 @@ class ZodbObjectHistory(object):
     def __getitem__(self, item):
         if self._history is None:
             self._load()
-        return self._history[item]
+        d = dict(self._history[item])
+        if isinstance(d['user_name'], bytes):
+            d['user_name'] = d['user_name'].decode('UTF-8', 'replace')
+        if isinstance(d['description'], bytes):
+            d['description'] = d['description'].decode('UTF-8', 'replace')
+        return d
 
     def lastChange(self, tid=None):
         if self._history is None:
@@ -95,15 +92,15 @@ class ZodbObjectHistory(object):
             self._obj._p_changed = True
 
 
+@adapter(IConnection)
+@implementer(IDatabaseHistory)
 class ZodbHistory(object):
 
-    adapts(IConnection)
-    implements(IDatabaseHistory)
-
     def __init__(self, connection):
         self._connection = connection
-        self._storage = connection._storage
+        self._storage = IStorageIteration(connection._storage)
         self._tids = cache.getStorageTids(self._storage)
+        self._iterators = []
 
     @property
     def tids(self):
@@ -112,12 +109,31 @@ class ZodbHistory(object):
     def __len__(self):
         return len(self._tids)
 
+    def _addcleanup(self, it):
+        self._iterators.append(it)
+        return it
+
+    def cleanup(self):
+        for it in self._iterators:
+            it.close()
+        self._iterators = []
+
     def __iter__(self):
-        return self._storage.iterator()
+        return self._addcleanup(self._storage.iterator())
 
-    def __getslice__(self, start, stop):
-        tids = self._tids[start:stop]
-        if not tids:
-            return []
-        return self._storage.iterator(tids[0], tids[-1])
+    def __getitem__(self, index):
+        if isinstance(index, slice):
+            assert index.step is None or index.step == 1
+            tids = self._tids[index]
+            if not tids:
+                return []
+            return self._addcleanup(self._storage.iterator(tids[0], tids[-1]))
+        else:
+            tid = self._tids[index]
+            return next(self._addcleanup(self._storage.iterator(tid, tid)))
 
+
+@adapter(MVCCAdapterInstance)
+@implementer(IStorageIteration)
+def getIterableStorage(storage):
+    return storage._storage
