@@ -14,6 +14,7 @@ import optparse
 import logging
 import errno
 import traceback
+import warnings
 
 from ZEO.ClientStorage import ClientStorage
 from ZODB.DB import DB
@@ -22,7 +23,7 @@ from ZODB.FileStorage.FileStorage import FileStorage
 from ZODB.interfaces import IDatabase
 from zope.app.server.servertype import IServerType
 from zope.app.appsetup.appsetup import SystemConfigurationParticipation
-from zope.component import getUtility, provideUtility
+from zope.component import getUtility, queryUtility, provideUtility
 from zope.server.taskthreads import ThreadedTaskDispatcher
 from zope.event import notify
 from zope.exceptions import exceptionformatter
@@ -32,11 +33,16 @@ from zodbbrowser.state import install_provides_hack
 
 
 class Options(object):
+    db_filename = None
+    zeo_address = None
+    zeo_storage = '1'
+    zeo_timeout = 30  # seconds
+    readonly = True
     listen_on = ('localhost', 8070)
     server_type = 'WSGI-HTTP'
     verbose = True
     threads = 4
-    features = ('zserver',) # maybe 'devmode' too?
+    features = ('zserver', 'standalone-zodbbrowser') # maybe 'devmode' too?
     site_definition = """
         <configure xmlns="http://namespaces.zope.org/zope"
                    i18n_domain="zodbbrowser">
@@ -91,7 +97,12 @@ def configure(options):
     xmlconfig.registerCommonDirectives(context)
     for feature in options.features:
         context.provideFeature(feature)
-    context = xmlconfig.string(options.site_definition, context=context)
+    with warnings.catch_warnings():
+        # zope.app.security globalmodules.zcml declares security directives for
+        # some modules deprecated in newer versions of Python.
+        warnings.filterwarnings('ignore',
+                                message='^the formatter module is deprecated')
+        context = xmlconfig.string(options.site_definition, context=context)
 
     endInteraction()
 
@@ -112,7 +123,7 @@ def start_server(options, db):
                                     port=port, db=db,
                                     task_dispatcher=task_dispatcher,
                                     verbose=options.verbose)
-    except socket.error, e:
+    except socket.error as e:
         if e.errno == errno.EADDRINUSE:
             sys.exit("Cannot listen on %s:%s: %s" % (host or '0.0.0.0',
                                                      port, e))
@@ -124,8 +135,8 @@ def start_server(options, db):
         port = server.socket.getsockname()[1]
 
     if options.verbose:
-        print "Listening on http://%s:%d/" % (host or socket.gethostname(),
-                                              port)
+        print("Listening on http://%s:%d/"
+              % (host or socket.gethostname(), port))
 
 
 def serve_forever(interval=30.0):
@@ -141,7 +152,9 @@ def stop_serving():
     task_dispatcher.shutdown(False)
     task_dispatcher = None
     asyncore.close_all()
-    getUtility(IDatabase, '<target>').close()
+    db = queryUtility(IDatabase, '<target>')
+    if db:
+        db.close()
 
 
 def monkeypatch_error_formatting():
@@ -150,9 +163,7 @@ def monkeypatch_error_formatting():
     traceback.print_exception = exceptionformatter.print_exception
 
 
-def main(args=None, start_serving=True):
-    logging.basicConfig(format="%(message)s")
-
+def parse_args(args=None):
     if args is None:
         args = sys.argv[1:]
 
@@ -208,8 +219,7 @@ def main(args=None, start_serving=True):
     monkeypatch_error_formatting()
 
     if opts.db:
-        filename = opts.db
-        db = DB(FileStorage(filename, read_only=opts.readonly))
+        options.db_filename = opts.db
     elif opts.zeo:
         if ':' in opts.zeo:
             # remote hostname:port ZEO connection
@@ -234,10 +244,34 @@ def main(args=None, start_serving=True):
             zeo_storage = opts.storage
         else:
             zeo_storage = '1'
-        db = DB(ClientStorage(zeo_address, storage=zeo_storage, read_only=opts.readonly))
+        options.zeo_address = zeo_address
+        options.zeo_storage = zeo_storage
     else:
         parser.error('please specify a database')
 
+    options.readonly = opts.readonly
+
+    return options
+
+
+def open_db(options):
+    if options.db_filename:
+        storage = FileStorage(options.db_filename, read_only=options.readonly)
+    else:
+        storage = ClientStorage(options.zeo_address,
+                                wait_timeout=options.zeo_timeout,
+                                storage=options.zeo_storage,
+                                read_only=options.readonly)
+    return DB(storage)
+
+
+def main(args=None, start_serving=True):
+    logging.basicConfig(format="%(message)s")
+
+    options = parse_args(args)
+
+    db = open_db(options)
+
     internal_db = DB(MappingStorage())
 
     configure(options)
