@@ -3,8 +3,8 @@ import logging
 from persistent.dict import PersistentDict
 from persistent.list import PersistentList
 from persistent.mapping import PersistentMapping
-from zope.component import adapts, getMultiAdapter
-from zope.interface import implements, Interface
+from zope.component import adapter, getMultiAdapter
+from zope.interface import implementer, Interface
 from zope.interface.interfaces import IInterface
 from zope.interface.interface import InterfaceClass
 from zope.proxy import removeAllProxies
@@ -28,6 +28,7 @@ except ImportError:
     from zope.app.container.contained import ContainedProxy # BBB
 
 from zodbbrowser.interfaces import IStateInterpreter, IObjectHistory
+from zodbbrowser.history import ZodbObjectHistory
 
 
 log = logging.getLogger(__name__)
@@ -42,12 +43,17 @@ def install_provides_hack():
     A common result of missing modules in sys.path is that you cannot
     unpickle objects that have been marked with directlyProvides() to
     implement interfaces that aren't currently available.  Those interfaces
-    are replaced by persistent broken placeholders, which aren classes,
+    are replaced by persistent broken placeholders, which are classes,
     not interfaces, and aren't iterable, causing TypeErrors during unpickling.
     """
     zope.interface.declarations.Provides = Provides
 
 
+def uninstall_provides_hack():
+    """Undo the monkey-patch installed by install_provides_hack()."""
+    zope.interface.declarations.Provides = real_Provides
+
+
 def flatten_interfaces(args):
     result = []
     for a in args:
@@ -65,19 +71,20 @@ def flatten_interfaces(args):
 def Provides(cls, *interfaces):
     try:
         return real_Provides(cls, *interfaces)
-    except TypeError, e:
+    except TypeError as e:
         log.warning('Suppressing TypeError while unpickling Provides: %s', e)
         args = flatten_interfaces(interfaces)
         return real_Provides(cls, *args)
 
 
+@implementer(IStateInterpreter)
 class ZodbObjectState(object):
-    implements(IStateInterpreter)
 
     def __init__(self, obj, tid=None, _history=None):
         self.obj = removeAllProxies(obj)
         if _history is None:
-            _history = IObjectHistory(self.obj)
+            # Not using IObjectHistory(self.obj) because LP#1185175
+            _history = ZodbObjectHistory(self.obj)
         else:
             assert _history._obj is self.obj
         self.history = _history
@@ -88,17 +95,16 @@ class ZodbObjectState(object):
         self._load()
 
     def _load(self):
-        self.tid = self.history.lastChange(self.requestedTid)
         try:
+            self.tid = self.history.lastChange(self.requestedTid)
             self.pickledState = self.history.loadStatePickle(self.tid)
             loadedState = self.history.loadState(self.tid)
-        except Exception, e:
-            self.loadError = "%s: %s" % (e.__class__.__name__, e)
-            self.state = LoadErrorState(self.loadError, self.requestedTid)
-        else:
             self.state = getMultiAdapter((self.obj, loadedState,
                                          self.requestedTid),
                                          IStateInterpreter)
+        except Exception as e:
+            self.loadError = "%s: %s" % (e.__class__.__name__, e)
+            self.state = LoadErrorState(self.loadError, self.requestedTid)
 
     def getError(self):
         return self.loadError
@@ -143,9 +149,9 @@ class ZodbObjectState(object):
             return ZodbObjectState(parent, self.requestedTid)
 
 
+@implementer(IStateInterpreter)
 class LoadErrorState(object):
     """Placeholder for when an object's state could not be loaded"""
-    implements(IStateInterpreter)
 
     def __init__(self, error, tid):
         self.error = error
@@ -170,10 +176,10 @@ class LoadErrorState(object):
         return {}
 
 
+@adapter(Interface, dict, None)
+@implementer(IStateInterpreter)
 class GenericState(object):
     """Most persistent objects represent their state as a dict."""
-    adapts(Interface, dict, None)
-    implements(IStateInterpreter)
 
     def __init__(self, type, state, tid):
         self.state = state
@@ -198,9 +204,9 @@ class GenericState(object):
         return self.state
 
 
+@adapter(PersistentMapping, dict, None)
 class PersistentMappingState(GenericState):
     """Convenient access to a persistent mapping's items."""
-    adapts(PersistentMapping, dict, None)
 
     def listItems(self):
         return sorted(self.state.get('data', {}).items())
@@ -216,19 +222,19 @@ if PersistentMapping is PersistentDict:
     class DecoyPersistentDict(PersistentMapping):
         """Decoy to avoid ZCML errors while supporting both ZODB 3.8 and 3.9."""
 
+    @adapter(DecoyPersistentDict, dict, None)
     class PersistentDictState(PersistentMappingState):
         """Decoy to avoid ZCML errors while supporting both ZODB 3.8 and 3.9."""
-        adapts(DecoyPersistentDict, dict, None)
 
-else:
+else:  # pragma: nocover
+    @adapter(PersistentDict, dict, None)
     class PersistentDictState(PersistentMappingState):
         """Convenient access to a persistent dict's items."""
-        adapts(PersistentDict, dict, None)
 
 
+@adapter(SampleContainer, dict, None)
 class SampleContainerState(GenericState):
     """Convenient access to a SampleContainer's items"""
-    adapts(SampleContainer, dict, None)
 
     def listItems(self):
         data = self.state.get('_SampleContainer__data')
@@ -243,9 +249,9 @@ class SampleContainerState(GenericState):
                                IStateInterpreter).listItems()
 
 
+@adapter(OrderedContainer, dict, None)
 class OrderedContainerState(GenericState):
     """Convenient access to an OrderedContainer's items"""
-    adapts(OrderedContainer, dict, None)
 
     def listItems(self):
         # Now this is tricky: we want to construct a small object graph using
@@ -265,8 +271,8 @@ class OrderedContainerState(GenericState):
         return container.items()
 
 
+@adapter(ContainedProxy, tuple, None)
 class ContainedProxyState(GenericState):
-    adapts(ContainedProxy, tuple, None)
 
     def __init__(self, proxy, state, tid):
         GenericState.__init__(self, proxy, state, tid)
@@ -290,10 +296,10 @@ class ContainedProxyState(GenericState):
         return dict(self.listAttributes())
 
 
+@adapter(Interface, Interface, None)
+@implementer(IStateInterpreter)
 class FallbackState(object):
     """Fallback when we've got no idea how to interpret the state"""
-    adapts(Interface, Interface, None)
-    implements(IStateInterpreter)
 
     def __init__(self, type, state, tid):
         self.state = state
