@@ -1,4 +1,6 @@
 import doctest
+import textwrap
+import unittest
 
 from zope.app.testing import setup
 from zope.component import provideAdapter
@@ -14,8 +16,8 @@ def pprintDict(d):
     pprint.pprint() doesn't cut it: when the dict is too short, it uses
     repr() which has no defined ordering.
     """
-    print '{%s}' % ',\n '.join('%r: %r' % (key, value)
-                               for key, value in sorted(d.items()))
+    print('{%s}' % ',\n '.join('%r: %r' % (key, value)
+                               for key, value in sorted(d.items())))
 
 
 def doctest_compareDicts():
@@ -72,7 +74,7 @@ def doctest_compareTuplesHTML():
 
         >>> old = (1, 2, 3, 5)
         >>> new = (2, 3, 4, 5)
-        >>> print compareTuplesHTML(new, old)
+        >>> print(compareTuplesHTML(new, old))
         <div class="diff">
           <div class="diffitem removed">
             removed 1
@@ -97,9 +99,9 @@ def doctest_compareTuplesHTML():
           </div>
         </div>
 
-        >>> old = (1, 2, 3, 5)
-        >>> new = (1, 3, 4, 5)
-        >>> print compareTuplesHTML(new, old)
+        >>> old = (1, 2, 3, 5, 6, 7)
+        >>> new = (1, 3, 4, 5, 6, 7)
+        >>> print(compareTuplesHTML(new, old))
         <div class="diff">
           <div class="diffitem same">
             first item kept the same
@@ -117,16 +119,16 @@ def doctest_compareTuplesHTML():
             added 4
           </div>
           <div class="diffitem same">
-            last item kept the same
+            last 3 items kept the same
           </div>
         </div>
 
-        >>> old = (1, 2, 3, 5)
-        >>> new = (1, 3, 4, 6)
-        >>> print compareTuplesHTML(new, old)
+        >>> old = (0, 1, 2, 3, 5)
+        >>> new = (0, 1, 3, 4, 6)
+        >>> print(compareTuplesHTML(new, old))
         <div class="diff">
           <div class="diffitem same">
-            first item kept the same
+            first 2 items kept the same
           </div>
           <div class="diffitem removed">
             removed 2
@@ -150,7 +152,7 @@ def doctest_compareTuplesHTML():
 
         >>> old = (1, 2, 3, 5)
         >>> new = (2, 3, 4, 6)
-        >>> print compareTuplesHTML(new, old)
+        >>> print(compareTuplesHTML(new, old))
         <div class="diff">
           <div class="diffitem removed">
             removed 1
@@ -186,7 +188,7 @@ def doctest_compareDictsHTML():
 
         >>> old = dict(a=1, b=2, c=3)
         >>> new = dict(a=1, b=3, e=4)
-        >>> print compareDictsHTML(new, old)
+        >>> print(compareDictsHTML(new, old))
         <div class="diff">
           <div class="diffitem changed">
             <strong>b</strong>: changed to 3
@@ -207,7 +209,7 @@ def doctest_compareDictsHTML_recursive():
 
         >>> old = dict(a=1, b=dict(x=2, y=5), c=3, d=42, e=dict(x=42))
         >>> new = dict(a=1, b=dict(x=3, y=5), d=dict(x=42), e=42, f=dict(x=4))
-        >>> print compareDictsHTML(new, old)
+        >>> print(compareDictsHTML(new, old))
         <div class="diff">
           <div class="diffitem changed">
             <strong>b</strong>: dictionary changed:
@@ -231,6 +233,35 @@ def doctest_compareDictsHTML_recursive():
           </div>
         </div>
 
+        >>> old = dict(a=1, b=(2, 5), c=3, d=42)
+        >>> new = dict(a=1, b=(3, 5), d=(42, ), e='x')
+        >>> print(compareDictsHTML(new, old))
+        <div class="diff">
+          <div class="diffitem changed">
+            <strong>b</strong>: tuple changed:
+            <div class="diff">
+              <div class="diffitem removed">
+                removed 2
+              </div>
+              <div class="diffitem added">
+                added 3
+              </div>
+              <div class="diffitem same">
+                last item kept the same
+              </div>
+            </div>
+          </div>
+          <div class="diffitem removed">
+            <strong>c</strong>: removed 3
+          </div>
+          <div class="diffitem changed">
+            <strong>d</strong>: changed to (42,)
+          </div>
+          <div class="diffitem added">
+            <strong>e</strong>: added 'x'
+          </div>
+        </div>
+
     """
 
 
@@ -239,7 +270,7 @@ def doctest_compareDictsHTML_tid_is_used():
 
         >>> old = dict(a=1, b=dict(x=2, y=5))
         >>> new = dict(a=2, b=dict(x=3, y=5))
-        >>> print compareDictsHTML(new, old, tid=42)
+        >>> print(compareDictsHTML(new, old, tid=42))
         <div class="diff">
           <div class="diffitem changed">
             <strong>a</strong>: changed to 2 [tid=42]
@@ -262,7 +293,7 @@ def doctest_compareDictsHTML_html_quoting():
 
         >>> old = {}
         >>> new = {'<': 'less than'}
-        >>> print compareDictsHTML(new, old)
+        >>> print(compareDictsHTML(new, old))
         <div class="diff">
           <div class="diffitem added">
             <strong>&lt;</strong>: added 'less than'
@@ -277,7 +308,7 @@ def doctest_compareDictsHTML_nonstring_keys():
 
         >>> old = {1: 2}
         >>> new = {1: 3}
-        >>> print compareDictsHTML(new, old)
+        >>> print(compareDictsHTML(new, old))
         <div class="diff">
           <div class="diffitem changed">
             <strong>1</strong>: changed to 3
@@ -287,22 +318,38 @@ def doctest_compareDictsHTML_nonstring_keys():
     """
 
 
-def doctest_compareDictsHTML_has_no_unicode_problems():
-    r"""Tests for compareDicts
+class TestCompareDictsHTML(unittest.TestCase):
 
-        >>> old = {u'\N{SNOWMAN}': 1, '\xFE': 2}
-        >>> new = {u'\N{SNOWMAN}': 2, '\xFE': 3}
-        >>> print compareDictsHTML(new, old)
-        <div class="diff">
-          <div class="diffitem changed">
-            <strong>'\xfe'</strong>: changed to 3
-          </div>
-          <div class="diffitem changed">
-            <strong>u'\u2603'</strong>: changed to 2
-          </div>
-        </div>
+    def setUp(self):
+        setup.placelessSetUp()
+        provideAdapter(SimpleValueRenderer)
 
-    """
+    def tearDown(self):
+        setup.placelessTearDown()
+
+    def test_no_unicode_problems(self):
+        old = {u'\N{SNOWMAN}': 1, b'\xFE': 2}
+        new = {u'\N{SNOWMAN}': 2, b'\xFE': 3}
+        diff = compareDictsHTML(new, old)
+        self.assertMultiLineEqual(diff, textwrap.dedent("""\
+            <div class="diff">
+              <div class="diffitem changed">
+                <strong>'\\xfe'</strong>: changed to 3
+              </div>
+              <div class="diffitem changed">
+                <strong>u'\\u2603'</strong>: changed to 2
+              </div>
+            </div>
+        """ if str is bytes else """\
+            <div class="diff">
+              <div class="diffitem changed">
+                <strong>b'\\xfe'</strong>: changed to 3
+              </div>
+              <div class="diffitem changed">
+                <strong>'\u2603'</strong>: changed to 2
+              </div>
+            </div>
+        """))
 
 
 def setUp(test):
@@ -316,6 +363,9 @@ def tearDown(test):
 
 def test_suite():
     optionflags = doctest.REPORT_NDIFF | doctest.NORMALIZE_WHITESPACE
-    return doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
-                                optionflags=optionflags)
+    return unittest.TestSuite([
+        unittest.defaultTestLoader.loadTestsFromName(__name__),
+        doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
+                             optionflags=optionflags),
+    ])
 
