@@ -1,14 +1,13 @@
-import unittest
-import sys
-
 import transaction
 from persistent import Persistent
 from persistent.dict import PersistentDict
 from zope.interface.verify import verifyObject
+from zope.component import provideAdapter
 
 from zodbbrowser.tests.realdb import RealDatabaseTest
-from zodbbrowser.history import ZodbObjectHistory
-from zodbbrowser.interfaces import IObjectHistory
+from zodbbrowser.history import ZodbObjectHistory, ZodbHistory
+from zodbbrowser.history import getIterableStorage
+from zodbbrowser.interfaces import IObjectHistory, IDatabaseHistory
 
 
 class PersistentObject(Persistent):
@@ -26,11 +25,13 @@ class TestFileStorage(RealDatabaseTest):
         obj = self.obj
         history = ZodbObjectHistory(obj)
         verifyObject(IObjectHistory, history)
-        self.assertEquals(len(history), 1)
+        self.assertEqual(len(history), 1)
         self.assertTrue('tid' in history[0])
         self.assertTrue('time' in history[0])
         self.assertTrue('user_name' in history[0])
         self.assertTrue('description' in history[0])
+        self.assertNotIsInstance(history[0]['user_name'], bytes)
+        self.assertNotIsInstance(history[0]['description'], bytes)
 
     def test_some_history(self):
         obj = self.obj
@@ -38,13 +39,12 @@ class TestFileStorage(RealDatabaseTest):
             obj[n] = n
             transaction.commit()
         history = ZodbObjectHistory(obj)
-        self.assertEquals(len(history), 11)
+        self.assertEqual(len(history), 11)
 
 
-class TestLoadState(RealDatabaseTest):
+class WorkloadMixin(object):
 
-    def setUp(self):
-        RealDatabaseTest.setUp(self)
+    def commitSomeStuff(self):
         root = self.conn.root()
         self.adam = root['adam'] = PersistentObject()
         transaction.commit()
@@ -59,26 +59,33 @@ class TestLoadState(RealDatabaseTest):
         self.adam.laptop = 'ThinkPad T61'
         transaction.commit()
 
+
+class TestLoadState(WorkloadMixin, RealDatabaseTest):
+
+    def setUp(self):
+        RealDatabaseTest.setUp(self)
+        self.commitSomeStuff()
+
     def test_latest_state(self):
         state = ZodbObjectHistory(self.adam).loadState()
-        self.assertEquals(state, dict(laptop='ThinkPad T61'))
+        self.assertEqual(state, dict(laptop='ThinkPad T61'))
 
     def test_exact_state(self):
         tid = ZodbObjectHistory(self.adam)[1]['tid']
         state = ZodbObjectHistory(self.adam).loadState(tid)
-        self.assertEquals(state, dict(laptop='ThinkPad T42'))
+        self.assertEqual(state, dict(laptop='ThinkPad T42'))
 
     def test_earlier_state(self):
         tid = ZodbObjectHistory(self.eve)[0]['tid']
         state = ZodbObjectHistory(self.adam).loadState(tid)
-        self.assertEquals(state, dict(laptop='ThinkPad T23'))
+        self.assertEqual(state, dict(laptop='ThinkPad T23'))
 
     def test_error_handling(self):
         tid = ZodbObjectHistory(self.adam)[-1]['tid']
         history = ZodbObjectHistory(self.eve)
         try:
             history.loadState(tid)
-        except KeyError, e:
+        except KeyError as e:
             self.assertTrue("did not exist in or before" in str(e))
         else:
             self.fail("did not raise")
@@ -86,17 +93,41 @@ class TestLoadState(RealDatabaseTest):
     def test_rollback_does_nothing(self):
         history = ZodbObjectHistory(self.adam)
         history.rollback(history.lastChange())
-        self.assertEquals(self.adam.laptop, 'ThinkPad T61')
+        self.assertEqual(self.adam.laptop, 'ThinkPad T61')
         self.assertFalse(self.adam._p_changed)
 
     def test_rollback(self):
         history = ZodbObjectHistory(self.adam)
         history.rollback(history[-2]['tid'])
-        self.assertEquals(self.adam.laptop, 'ThinkPad T23')
+        self.assertEqual(self.adam.laptop, 'ThinkPad T23')
         self.assertTrue(self.adam._p_changed)
 
 
-def test_suite():
-    this = sys.modules[__name__]
-    return unittest.defaultTestLoader.loadTestsFromModule(this)
+class TestZodbHistory(WorkloadMixin, RealDatabaseTest):
+
+    def setUp(self):
+        super(TestZodbHistory, self).setUp()
+        provideAdapter(getIterableStorage)
+
+    def test_no_history(self):
+        history = ZodbHistory(self.conn)
+        self.addCleanup(history.cleanup)
+        verifyObject(IDatabaseHistory, history)
+        self.assertEqual(len(history), 1)
+        self.assertEqual(len(history.tids), 1)
+        self.assertEqual(len(list(history)), 1)
+        self.assertEqual([tr.tid for tr in history],
+                         [tr.tid for tr in history[-5:]])
+        self.assertEqual(history[10:], [])
+        self.assertEqual(history[0].tid, history.tids[0])
+
+    def test_some_history(self):
+        self.commitSomeStuff()
+        history = ZodbHistory(self.conn)
+        self.addCleanup(history.cleanup)
+        verifyObject(IDatabaseHistory, history)
+        self.assertEqual(len(history), 7)
+        self.assertEqual(len(list(history)), 7)
+        self.assertEqual([tr.tid for tr in history][-5:],
+                         [tr.tid for tr in history[-5:]])
 
