@@ -2,20 +2,33 @@ import logging
 import itertools
 import collections
 import re
-from cgi import escape
+from functools import partial
 
 from ZODB.utils import u64, oid_repr
 from persistent import Persistent
 from persistent.dict import PersistentDict
 from persistent.list import PersistentList
 from persistent.mapping import PersistentMapping
-from zope.component import adapts
+from zope.component import adapter
 from zope.interface.declarations import ProvidesClass
-from zope.interface import implements, Interface
+from zope.interface import implementer, Interface
 from zope.security.proxy import removeSecurityProxy
 
-from zodbbrowser.interfaces import IValueRenderer
-from zodbbrowser.interfaces import IObjectHistory
+
+from zodbbrowser.compat import basestring, escape
+from zodbbrowser.interfaces import IValueRenderer, IObjectHistory
+
+
+# Persistent has a __repr__ now that shows the OID, but it's shown poorly
+# (as a Python string full of backslash escapes) and includes lots of
+# irrelevant details (in-memory object address, repr of the Connection object).
+CLASSES_WITH_BAD_REPR = (object, Persistent)
+try:
+    from BTrees._base import Bucket, Set, Tree, TreeSet
+except ImportError:
+    pass
+else:
+    CLASSES_WITH_BAD_REPR += (Bucket, Set, Tree, TreeSet)
 
 
 log = logging.getLogger(__name__)
@@ -24,12 +37,12 @@ log = logging.getLogger(__name__)
 MAX_CACHE_SIZE = 1000
 TRUNCATIONS = {}
 TRUNCATIONS_IN_ORDER = collections.deque()
-next_id = itertools.count(1).next
+next_id = partial(next, itertools.count(1))
 
 
 def resetTruncations(): # for tests only!
     global next_id
-    next_id = itertools.count(1).next
+    next_id = partial(next, itertools.count(1))
     TRUNCATIONS.clear()
     TRUNCATIONS_IN_ORDER.clear()
 
@@ -46,20 +59,34 @@ def truncate(text):
     return id
 
 
+@adapter(Interface)
+@implementer(IValueRenderer)
 class GenericValue(object):
     """Default value renderer.
 
     Uses the object's __repr__, truncating if too long.
     """
-    adapts(Interface)
-    implements(IValueRenderer)
 
     def __init__(self, context):
         self.context = context
 
+    if hasattr(object.__repr__, '__func__'):  # pragma: nocover
+        # PyPy
+        def _same_method(self, a, b):
+            return getattr(a, '__func__', None) is b.__func__
+    else:
+        # CPython
+        def _same_method(self, a, b):
+            return a is b
+
+    def _has_no_repr(self, obj):
+        obj_repr = getattr(obj.__class__, '__repr__', None)
+        return any(self._same_method(obj_repr, cls.__repr__)
+                   for cls in CLASSES_WITH_BAD_REPR)
+
     def _repr(self):
         # hook for subclasses
-        if getattr(self.context.__class__, '__repr__', None) is object.__repr__:
+        if self._has_no_repr(self.context):
             # Special-case objects with the default __repr__ (LP#1087138)
             if isinstance(self.context, Persistent):
                 return '<%s.%s with oid %s>' % (
@@ -79,10 +106,10 @@ class GenericValue(object):
         if len(text) > limit:
             id = truncate(text[limit:])
             text = '%s<span id="%s" class="truncated">...</span>' % (
-                escape(text[:limit]), id)
+                escape(text[:limit], False), id)
         else:
-            text = escape(text)
-        if not isinstance(self.context, basestring):
+            text = escape(text, False)
+        if not isinstance(self.context, (basestring, bytes)):
             try:
                 n = len(self.context)
             except Exception:
@@ -112,29 +139,32 @@ def join_with_commas(html, open, close):
     return prefix + '<br />'.join(html) + suffix
 
 
+@adapter(basestring)
+@implementer(IValueRenderer)
 class StringValue(GenericValue):
     """String renderer."""
-    adapts(basestring)
-    implements(IValueRenderer)
 
     def __init__(self, context):
         self.context = context
 
     def render(self, tid=None, can_link=True, limit=200, threshold=4):
-        if self.context.count('\n') <= threshold:
+        newline = b'\n' if isinstance(self.context, bytes) else u'\n'
+        if self.context.count(newline) <= threshold:
             return GenericValue.render(self, tid, can_link=can_link,
                                        limit=limit)
         else:
-            if isinstance(self.context, unicode):
-                prefix = 'u'
-                context = self.context
-            else:
-                prefix = ''
+            if isinstance(self.context, bytes):
                 context = self.context.decode('latin-1').encode('ascii',
                                                             'backslashreplace')
+            else:
+                context = self.context
+            if isinstance(self.context, str):
+                prefix = ''
+            else:
+                prefix = 'u'
             lines = [re.sub(r'^[ \t]+',
                             lambda m: '&nbsp;' * len(m.group(0).expandtabs()),
-                            escape(line))
+                            escape(line, False))
                      for line in context.splitlines()]
             nl = '<br />' # hm, maybe '\\n<br />'?
             if sum(map(len, lines)) > limit:
@@ -149,10 +179,10 @@ class StringValue(GenericValue):
                         + "'</span>")
 
 
+@adapter(tuple)
+@implementer(IValueRenderer)
 class TupleValue(object):
     """Tuple renderer."""
-    adapts(tuple)
-    implements(IValueRenderer)
 
     def __init__(self, context):
         self.context = context
@@ -172,10 +202,10 @@ class TupleValue(object):
         return result
 
 
+@adapter(list)
+@implementer(IValueRenderer)
 class ListValue(object):
     """List renderer."""
-    adapts(list)
-    implements(IValueRenderer)
 
     def __init__(self, context):
         self.context = context
@@ -190,10 +220,10 @@ class ListValue(object):
         return result
 
 
+@adapter(dict)
+@implementer(IValueRenderer)
 class DictValue(object):
     """Dict renderer."""
-    adapts(dict)
-    implements(IValueRenderer)
 
     def __init__(self, context):
         self.context = context
@@ -204,19 +234,19 @@ class DictValue(object):
             html.append(IValueRenderer(key).render(tid, can_link) + ': ' +
                         IValueRenderer(value).render(tid, can_link))
         if (sum(map(len, html)) < threshold and
-            '<span class="struct">' not in ''.join(html)):
+                '<span class="struct">' not in ''.join(html)):
             return '{%s}' % ', '.join(html)
         else:
             return join_with_commas(html, '{', '}')
 
 
+@adapter(Persistent)
+@implementer(IValueRenderer)
 class PersistentValue(object):
     """Persistent object renderer.
 
     Uses __repr__ and makes it a hyperlink to the actual object.
     """
-    adapts(Persistent)
-    implements(IValueRenderer)
 
     view_name = '@@zodbbrowser'
     delegate_to = GenericValue
@@ -228,7 +258,7 @@ class PersistentValue(object):
         obj = self.context
         url = '%s?oid=0x%x' % (self.view_name, u64(self.context._p_oid))
         if tid is not None:
-            url += "&tid=%d" % u64(tid)
+            url += "&tid=0x%x" % u64(tid)
             try:
                 oldstate = IObjectHistory(self.context).loadState(tid)
                 clone = self.context.__class__.__new__(self.context.__class__)
@@ -240,18 +270,19 @@ class PersistentValue(object):
                           self.context.__class__, u64(self.context._p_oid))
         value = self.delegate_to(obj).render(tid, can_link=False)
         if can_link:
-            return '<a class="objlink" href="%s">%s</a>' % (escape(url), value)
+            return '<a class="objlink" href="%s">%s</a>' % (escape(url, True),
+                                                            value)
         else:
             return value
 
 
+@adapter(PersistentMapping)
 class PersistentMappingValue(PersistentValue):
-    adapts(PersistentMapping)
     delegate_to = DictValue
 
 
+@adapter(PersistentList)
 class PersistentListValue(PersistentValue):
-    adapts(PersistentList)
     delegate_to = ListValue
 
 
@@ -265,24 +296,24 @@ if PersistentMapping is PersistentDict:
     class DecoyPersistentDict(PersistentMapping):
         """Decoy to avoid ZCML errors while supporting both ZODB 3.8 and 3.9."""
 
+    @adapter(DecoyPersistentDict)
     class PersistentDictValue(PersistentValue):
         """Decoy to avoid ZCML errors while supporting both ZODB 3.8 and 3.9."""
-        adapts(DecoyPersistentDict)
         delegate_to = DictValue
 
-else:
+else:  # pragma: nocover
+    @adapter(PersistentDict)
     class PersistentDictValue(PersistentValue):
-        adapts(PersistentDict)
         delegate_to = DictValue
 
 
+@adapter(ProvidesClass)
+@implementer(IValueRenderer)
 class ProvidesValue(GenericValue):
     """zope.interface.Provides object renderer.
 
     The __repr__ of zope.interface.Provides is decidedly unhelpful.
     """
-    adapts(ProvidesClass)
-    implements(IValueRenderer)
 
     def _repr(self):
         return '<Provides: %s>' % ', '.join(i.__identifier__
